/*
 *   Copyright Institute of Communication and Computer Systems (ICCS)
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */

/**
 * @file   src/initialize.c
 * @author Ioannis Koutras (joko@microlab.ntua.gr)
 * @date   December 2012
 * @brief  System allocator's global object initializer.
 */

#include <fcntl.h> /* for open() */
#include <sys/mman.h> /* for mmap() */

#include <dmmlib/config.h>
#include "dmmlib/dmmlib.h"
#include "parse_env.h"
#include "tls_allocator.h"

__thread allocator_t * tls_allocator = NULL;

locking_allocator_list_t destructed_list =
{ SLIST_HEAD_INITIALIZER(head), STATE_UNLOCKED };

/** Initializes space to store the allocator state in heap. */
allocator_t * initialize_allocator(void) {

    int fd = open("/dev/zero", O_RDWR);
    allocator_t *allocator = mmap(NULL, SYS_PAGESIZE,
            PROT_READ | PROT_WRITE, MAP_PRIVATE, fd, 0);

    if(__builtin_expect(!!(allocator == MAP_FAILED), 0)) {
        return NULL;
    }

    for(int i=0; i < 32; i++) {
        SLIST_INIT(&allocator->slab_rb_head[i]);
    }
    SLIST_INIT(&allocator->rb_head);
#ifdef HAVE_LOCKS
    INIT_DEALY_LOCK(allocator->edit_lock);
#endif /* HAVE_LOCKS */
#ifdef WITH_KNOBS
    allocator->dmm_knobs.sys_alloc_size = SYS_ALLOC_SIZE;
#ifdef GOOD_FIT
    allocator->dmm_knobs.fit_percentage = GOOD_FIT_PERCENTAGE;
#endif /* GOOD_FIT */
#ifdef COALESCING_VARIABLE
    allocator->dmm_knobs.max_coal_size = MAX_COAL_SIZE;
#endif /* COALESCING_VARIABLE */
#ifdef SPLITTING_VARIABLE
    allocator->dmm_knobs.min_split_size = MIN_SPLIT_SIZE;
#endif /* SPLITTING_VARIABLE */
#endif /* WITH_KNOBS */
#ifdef WITH_ALLOCATOR_STATS
#ifdef REQUEST_SIZE_INFO
    allocator->dmm_stats.total_mem_requested = 0;
#endif /* REQUEST_SIZE_INFO */
    allocator->dmm_stats.total_mem_allocated = 0;
    allocator->dmm_stats.live_objects = 0;
    allocator->dmm_stats.num_malloc = 0;
    allocator->dmm_stats.num_free = 0;
    allocator->dmm_stats.num_realloc= 0;
#ifdef WITH_MEMALIGN
    allocator->dmm_stats.num_memalign = 0;
#endif /* WITH_MEMALIGN */
#endif /* WITH_ALLOCATOR_STATS */
#ifdef PARSE_ENV
#if defined  WITH_MEM_TRACE || defined WITH_STATS_TRACE || defined WITH_DEBUG
    allocator->initialized = false;
#endif /* WITH_MEM_TRACE || WITH_STATS_TRACE || WITH_DEBUG */
#endif /* PARSE_ENV */

#ifdef PARSE_ENV
    parse_env(allocator);
#if defined  WITH_MEM_TRACE || defined WITH_STATS_TRACE || defined WITH_DEBUG
    allocator->initialized = true;
#endif /* WITH_MEM_TRACE || WITH_STATS_TRACE || WITH_DEBUG */
#endif /* PARSE_ENV */

    return allocator;
}
