/*
 *   Copyright Institute of Communication and Computer Systems (ICCS) 
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */

/**
 * @file   slab/malloc.c
 * @author Iraklis Anagnostopoulos
 * @date   March 2013
 * @brief  malloc() implementation for slab allocation
 */


#include <inttypes.h>

#include "dmmlib/raw_block.h"
#include "slab.h"

/** Tries to allocate memory from a specific free-list organized raw block.
 * @param slab_raw_block The pointer of the freelist-organised slab raw block.
 * @param size      The requested size.
 * @retval          The address of the returned memory space.
 */
void * slab_malloc(slab_rb_t *slab_raw_block, size_t size) {
    slab_block_header_t *ptr = NULL;
    raw_block_header_t *current_slab_raw_block;

    if((&slab_raw_block->fl_head)->slh_first != NULL) {
        if((&slab_raw_block->fl_head)->slh_first == slab_raw_block->fl_tail) {
            current_slab_raw_block = (raw_block_header_t *)
                ((uintptr_t) slab_raw_block - sizeof(raw_block_header_t));
            DEALY_LOCK(current_slab_raw_block->lock);
            ptr = (slab_block_header_t *) (&slab_raw_block->fl_head)->slh_first;
            SLIST_REMOVE_HEAD(&slab_raw_block->fl_head, pointers);
            slab_raw_block->fl_tail = (&slab_raw_block->fl_head)->slh_first;
            DEALY_UNLOCK(current_slab_raw_block->lock);
        } else {
            ptr = (slab_block_header_t *) (&slab_raw_block->fl_head)->slh_first;
            SLIST_REMOVE_HEAD(&slab_raw_block->fl_head, pointers);
        }
    } else {
        if(size <= slab_raw_block->remaining_size) {
            if(slab_raw_block->border_ptr == NULL) {
                ptr = (slab_block_header_t *)((uintptr_t) slab_raw_block +
                        sizeof(slab_rb_t));
            } else {
                ptr = (slab_block_header_t *)
                    ((uintptr_t) slab_raw_block->border_ptr + size);
            }

            // Update raw block metadata
            slab_raw_block->remaining_size -= size;
            slab_raw_block->border_ptr = ptr;
        }
    }

    if(ptr != NULL) {
        return (void *) ptr;
    } else {
        return NULL;
    }
}
